#import <Foundation/Foundation.h>
#import "LogUtil.h"
#import "BlockAdditions.h"
#import "ExceptionAdditions.h"

#define YesNumber [NSNumber numberWithBool:YES]
#define NoNumber [NSNumber numberWithBool:NO]

#define Float(exp) [NSNumber numberWithFloat:exp]
#define Long(exp) [NSNumber numberWithLong:exp]
#define Int(exp) [NSNumber numberWithInt:exp]
#define Bool(exp) [NSNumber numberWithBool:exp]
#define Double(exp) [NSNumber numberWithDouble:exp]

#define Fmt(...) [NSString stringWithFormat:__VA_ARGS__]

#define PropertyAffected(CapitalisedProperty, ...) + (NSSet*) keyPathsForValuesAffecting##CapitalisedProperty { \
return [NSSet setWithObjects:__VA_ARGS__, nil]; \
}

@interface NSObject (UsefulAdditions)

- (NSString*) classname;
#if TARGET_OS_IPHONE
- (NSString*) className;
#endif
- (void) performSelector:(SEL) sel withArgumentPointer:(void*) arg;
- (NSArray*) singletonArray;
- (NSArray*) singletonArrayIfNotArray;

@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSNumber (UsefulAdditions)
// Note these are all based on objcType and are therefore unreliable according to the NSNumber docs.
// I have found this to be the case when using valueForKey on an object to get a property value - for BOOL
// properties it indicates int!
- (BOOL) isBoolean;
- (BOOL) isShort;
- (BOOL) isInt;
- (BOOL) isLong;
- (BOOL) isLongLong;
- (BOOL) isFloat;
- (BOOL) isDouble;
@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSString (UsefulAdditions)

+ (NSString*) stringWithData:(NSData*) data;
+ (BOOL) isBlank:(NSString*) s;
@property (readonly) BOOL notBlank; // negative is good because it returns the right answer when nil

- (BOOL) eq:(NSString*) other;
- (NSString*) app:(NSString*) other;
- (NSString*) appPath:(NSString*) other;
- (NSString*) trim;
- (NSString*) trimLeading:(NSString*) before andTrailing:(NSString*) after;
- (NSString*) lowercaseFirst;
- (NSString*) uppercaseFirst;
- (NSString*) escapedUrlParameter;
- (NSString*) escapedUrlPath;
- (NSArray*) split:(NSString*) delimiter;
- (int) indexOf:(NSString*) s;
- (int) indexOf:(NSString*) s from:(int) from;
- (int) indexAfter:(NSString*) s;
- (int) indexAfter:(NSString*) s  from:(int) from;
- (int) lastIndexOf:(NSString*) s;
- (BOOL) contains:(NSString*) s;
- (NSRange) rangeEnclosedBy:(NSString*) start and:(NSString*) end;
- (NSRange) rangeEnclosedBy:(NSString*) start and:(NSString*) end from:(int) from;
- (NSString*) substringFrom:(int) from to:(int) to;
- (NSString*) substringToFirstOccurrenceOf:(NSString*) s;
#if TARGET_OS_IPHONE
#else
- (BOOL) matchesComplete:(NSString *) exp;
- (BOOL) matchesPartial:(NSString *) exp;
- (BOOL) isValidEmail;
#endif
- (NSString*) emailDomain; // assumes it's a valid email address
- (NSString*) emailName; // ditto
- (BOOL) isValidInt;
- (BOOL) isValidByte; // 0-255
- (BOOL) isValidIpAddress;
- (BOOL) parseIpAddress:(int*) bytes;
- (BOOL) writeToFile:(NSString*) file;
- (NSString*) plural:(int) number;
- (NSString*) plural:(int) number suffix:(NSString*) suffix;
- (NSNumber*) floatNumber;
- (NSNumber*) intNumber;
- (NSString*) stringByAddingPercentEscapes;
- (NSString*) stringByAddingPercentEscapesWithExtras:(NSString*) extras;

// returns "pages.zip" for "marion.pages.zip", whereas pathExtension would return "zip"
- (NSString*) pathExtensionAllowingMultiple;

@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSArray (UsefulAdditions)

- (id) objectAtIndexOrNil:(int) i;

- (NSMutableArray*) arrayByRemovingObject:(NSObject*) ob;

// NSNulls are ok
- (void) makeObjectsPerformSelectorNicely:(SEL) sel withObject:(id) obj;

// NSNulls are ok
- (void) makeObjectsPerformSelectorNicely:(SEL) sel withArgumentPointer:(void*) arg;

- (id) firstObjectWithValue:(id) value forSelector:(SEL) selector;
- (id) firstObjectWithValue:(id) value forKeyPath:(NSString*) path;
- (NSArray*) arrayByMakingObjectsPerformSelector:(SEL) sel;
- (NSArray*) arrayByPassingObjectsToSelector:(SEL) sel ofObject:(NSObject*) ob;
- (NSArray*) arrayOfObjectsMatchingCall:(NSInvocation*) booleanCall invert:(BOOL) invert;

// Like a set, but the objects stay in the order in which they first appear in the array
- (NSArray*) uniqueObjects;

- (NSString*) join:(NSString*) joiner;
- (NSArray*) subarrayFromIndex:(int) index;
- (NSArray*) subarrayFromIndex:(int) index toIndex:(int) to;
@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSMutableArray (UsefulAdditions)
// Assuming the array is sorted, adds an object into the array at the appropriate place
- (int) addSorted:(id) object;
- (int) addSorted:(id) object compareProperty:(NSString*) compareProperty;
@end


#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSURL (UsefulAdditions)
+ (NSURL*) urlWithString:(NSString*) mainUrl parameters:(NSString*) parameters, ...;
- (NSURL*) urlByChangingSchemeTo:(NSString*) protocol;

// everything, including query parameters, from the file onwards (not decoded, note).
- (NSString*) relativePortion;

/**
 Data suitable for writing to a webloc file. bookmarkDataWithOptions doesn't exist until
 10.6, and is unimplemented (but defined) in iOS.
 */
- (NSData*) bookmarkData;

- (NSURL*) append:(NSString*) s;

@end


#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSInvocation (UsefulAdditions)

+ (NSInvocation*) invocationWithTarget:(id) target selector:(SEL) selector;
- (void) invokeWithArgument:(id) arg;
- (void) invokeWithBoolean:(BOOL) b;
- (void) invokeWithFloat:(float) f;

@end


#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSFileManager(UsefulAdditions)

- (BOOL) isDirectoryAtPath:(NSString*) path;

@end


#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSDictionary (UsefulAdditions)
- (NSDictionary*) dictionaryWithPropertiesOfObject:(NSObject*) o;
- (NSDictionary*) dictionaryByPassingValuesToSelector:(SEL) sel ofObject:(NSObject*) ob;

- (int) integerForKey:(id) key;
- (int) integerForKey:(id) key defaultValue: (int) defaultValue;
- (BOOL) boolForKey:(id) key;
- (BOOL) boolForKey:(id) key defaultValue: (BOOL) defaultValue;

- (BOOL) containsKey:(id) key;
@end


#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSMutableDictionary(UsefulAdditions)

- (void) setBool:(BOOL)value forKey:(id) key;
- (void) setInteger:(int)value forKey:(id) key;

@end


#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSUserDefaults (UsefulAdditions)

- (NSMutableDictionary*) mutableDictionaryForKey:(NSString*) key;
- (BOOL) boolForKey:(NSString*) key defaultingTo:(BOOL) def;
- (NSString*) stringForKey:(NSString*) key defaultingTo:(NSString*) def;
- (int) integerForKey:(NSString*) key defaultingTo:(int) def;
- (id) objectForKey:(NSString *)key defaultingTo:(id) def;

@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface Swizzler : NSObject

+ (BOOL) swizzle:(Class) klass method:(SEL)origSel_ withMethod:(SEL)altSel_;
+ (BOOL) swizzleClassMethod:(Class) klass method:(SEL)origSel_ withMethod:(SEL)altSel_;

@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSDate (UsefulAdditions)

- (BOOL) before:(NSDate*) other;
- (BOOL) after:(NSDate*) other;
- (NSString*) ymd;
- (NSString*) ymdDash;
- (NSDate*) dateByAddingDays:(int) days;
- (NSDate*) dateByAddingMonths:(int) months;
- (NSDate*) dateByAddingYears:(int) years;
- (NSDate*) midnight;
- (NSString*) ymdhmDash;

@end

#pragma mark -------------------------------------------------------------------------------------------------------------------

@interface NSDateFormatter (UsefulAdditions)

+ (NSDate*) dateFromEnglishString:(NSString*) dateString withFormat:(NSString*) format;
+ (NSDate*) dateFromLocalLanguageString:(NSString*) dateString withFormat:(NSString*) format;
+ (NSDate*) dateFromString:(NSString*) dateString withFormat:(NSString*) format localeString:(NSString*) localeString;

@end
